#!/usr/bin/env python

import itertools

import networkx as nx
import numpy as np


def display_PPI_network_for_complex(ax, bp_PPI_df, Corum_DF, Complex_ID, 
    node_size, edge_width, node_font_size = 10, bait_node_color='xkcd:red', 
    prey_node_color='xkcd:rose pink', AP_MS_edge_color='xkcd:red',
    node_pos=False):
    '''
    Display network of BioPlex PPIs for a CORUM complex.
    
    This function displays a complete network in which nodes represent 
    the proteins in a specified CORUM complex and edges represent 
    BioPlex PPIs using NetworkX. Edges detected through AP-MS are
    colored darker.

    Parameters
    ----------
    ax object to draw on: Matplotlib Axes
    DataFrame of PPIs : Pandas DataFrame
    DataFrame of CORUM complexes : Pandas DataFrame
    Corum Complex ID: int
    Size of Nodes in Network: int
    Width of Edges in Network: float
    Size of font for Node Labels: int (optional)
    Color of Nodes targeted as baits: str (optional)
    Color of Nodes detected as preys only: str (optional)
    Color of Edges observed via AP-MS from PPI data: str (optional)
    Networkx Position of Nodes: dict (optional)

    Returns
    -------
    Node Positions
        Dictionary of Node Positions in NetworkX layout

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    # (2) Obtain CORUM complexes
    # (3) create figure and axis objects to draw on
    # (4) Visualize network for specified protein complex
    #     using PPI data (ING2 complex ID: 2851)
    >>> bp_PPI_df = getBioPlex('293T', '3.0')
    >>> Corum_DF = getCorum()
    >>> fig, ax = plt.subplots()
    >>> ING2_node_layout = display_PPI_network_for_complex(ax, bp_PPI_df, Corum_DF, 2851, 2300, 3.5)
    >>> ING2_node_layout
    {'A0A024R3R1': array([ 1.00000000e+00, -2.29248628e-09]), 'O75446': array([0.88545603, 0.46472316]), 'P29374': array([0.56806476, 0.82298384]), 'Q09028': array([0.12053671, 0.99270884]), 'Q13547': array([-0.35460481,  0.93501625]), 'Q16576': array([-0.74851068,  0.66312264]), 'Q5PSV4': array([-0.9709418 ,  0.23931567]), 'Q92769': array([-0.9709418 , -0.23931561]), 'Q96ST3': array([-0.7485108 , -0.66312258]), 'Q9H0E3': array([-0.35460499, -0.9350162 ]), 'Q9H160': array([ 0.12053676, -0.99270884]), 'Q9H7L9': array([ 0.56806458, -0.82298396]), 'Q9HCU9': array([ 0.88545603, -0.46472319])}
    '''
    # store uniprot IDs & gene symbols that belong to this complex in a list
    genes_in_complex_i = (Corum_DF[Corum_DF.complex_id == Complex_ID]
                            .loc[:,'subunits_uniprot_id'].values[0]
                            .split(';')) # Uniprot
    
    gene_symbols_in_complex_i = (Corum_DF[Corum_DF.complex_id == Complex_ID]
                                 .loc[:,'subunits_gene_name'].values[0]
                                 .split(';')) # Symbol

    # filter BioPlex PPI dataframe to include only interactions 
    # where both genes are found in complex
    complex_i_PPI_filter = []
    for uniprot_A, uniprot_B in zip(bp_PPI_df.UniprotA, bp_PPI_df.UniprotB):
        
        # check for isoform IDs and adjust
        if '-' in uniprot_A:
            uniprot_A = uniprot_A.split('-')[0]
        if '-' in uniprot_B:
            uniprot_B = uniprot_B.split('-')[0]

        # check to see if both gene symbols for this 
        # interaction are genes in complex
        if ((uniprot_A in genes_in_complex_i) and 
            (uniprot_B in genes_in_complex_i)):
            complex_i_PPI_filter.append(True)
        else:
            complex_i_PPI_filter.append(False)

    complex_i_PPI_filter = np.array(complex_i_PPI_filter)
    # use filter to subset bp PPI dataframe
    bp_complex_i_df = bp_PPI_df[complex_i_PPI_filter]
    bp_complex_i_df.reset_index(inplace = True, drop = True) # reset index

    # reconstruct UniprotA/UniprotB columns without '-' isoform id
    UniprotA_new = []
    UniprotB_new = []
    for UniprotA, UniprotB in zip(bp_complex_i_df.UniprotA,
                                  bp_complex_i_df.UniprotB):

        if '-' in UniprotA:
            UniprotA_new.append(UniprotA.split('-')[0])
        else:
            UniprotA_new.append(UniprotA)

        if '-' in UniprotB:
            UniprotB_new.append(UniprotB.split('-')[0])
        else:
            UniprotB_new.append(UniprotB)
            
    # update columns for Uniprot source & 
    # Uniprot target to exclude isoform '-' ID
    bp_complex_i_df.loc[:,'UniprotA'] = UniprotA_new
    bp_complex_i_df.loc[:,'UniprotB'] = UniprotB_new
    
    # subset PPI dataframe to the cols we need to construct graph
    bp_complex_i_df = (bp_complex_i_df.loc[:,
                        ['UniprotA','UniprotB','SymbolA','SymbolB']])
    
    # create a graph from the nodes/genes of specified complex
    bp_complex_i_G = nx.Graph()
    bp_complex_i_G.add_nodes_from(genes_in_complex_i)
    
    # iterate over AP-MS interactions in PPI df and add edges
    for source, target in zip(bp_complex_i_df.UniprotA, 
                              bp_complex_i_df.UniprotB):
        bp_complex_i_G.add_edge(source, target)
        
    # get mapping uniprot -> symbol & store as node attribute 
    # from CORUM complex data
    uniprot_symbol_dict = dict([key,val] for key, val in 
                            zip(genes_in_complex_i, gene_symbols_in_complex_i))
    for node_i in bp_complex_i_G.nodes():
        bp_complex_i_G.nodes[node_i]["symbol"] = uniprot_symbol_dict[node_i]

    # get a list of genes that were identifed as "baits" 
    # and "preys" for coloring nodes
    bp_complex_i_baits = list(set(bp_complex_i_df.UniprotA))
    bp_complex_i_preys = list(set(bp_complex_i_df.UniprotB))

    labels = dict([(key,val) for key, val in zip(list(bp_complex_i_G.nodes), 
                [bp_complex_i_G.nodes[node_i]['symbol'] for 
                node_i in bp_complex_i_G.nodes])])

    # color nodes according to whether they were present among "baits" 
    # & "preys", just "baits", just "preys" or not detected in PPI data
    # for this complex
    node_color_map = []
    for node_i_uniprot in bp_complex_i_G.nodes:

        # gene is present among baits & preys for the PPIs 
        # detected in this complex
        if ((node_i_uniprot in bp_complex_i_baits) and 
            (node_i_uniprot in bp_complex_i_preys)):
            node_color_map.append(bait_node_color)

        # gene is present among baits but NOT preys for 
        # the PPIs detected in this complex
        elif ((node_i_uniprot in bp_complex_i_baits) and 
              (node_i_uniprot not in bp_complex_i_preys)):
            node_color_map.append(bait_node_color)

        # gene is NOT present among baits but is present among 
        # preys for the PPIs detected in this complex
        elif ((node_i_uniprot not in bp_complex_i_baits) and 
              (node_i_uniprot in bp_complex_i_preys)):
            node_color_map.append(prey_node_color)

        # gene is NOT present among baits and is NOT present among 
        # preys for the PPIs detected in this complex
        elif ((node_i_uniprot not in bp_complex_i_baits) and 
              (node_i_uniprot not in bp_complex_i_preys)):
            node_color_map.append('0.7')

    # create a complete graph from the nodes of complex graph to add in all 
    # "background" edges (edges detected with AP-MS will be colored over)
    # position will be the same for both graphs since nodes are the same
    bp_complex_i_G_complete = nx.Graph()
    bp_complex_i_G_complete.add_nodes_from(bp_complex_i_G.nodes)
    bp_complex_i_G_complete.add_edges_from(
        itertools.combinations(bp_complex_i_G.nodes, 2))

    # optional argument "node_pos" used here
    # check to see if node position object has been fed as an argument
    if node_pos == False:
        # DEFAULT: set position of nodes w/ circular layout
        pos = nx.circular_layout(bp_complex_i_G)
    else:
        pos = node_pos # use node positions fed into function

    # construct edges for COMPLETE graph for "background" edges
    edges_complete = nx.draw_networkx_edges(bp_complex_i_G_complete, 
                                            pos, width = edge_width, 
                                            alpha = 0.25, ax = ax)
    edges_complete.set_edgecolor("xkcd:grey")

    # construct edges
    edges = nx.draw_networkx_edges(bp_complex_i_G, pos, 
                                   width = edge_width, ax = ax)
    edges.set_edgecolor(AP_MS_edge_color)

    # construct nodes
    nodes = nx.draw_networkx_nodes(bp_complex_i_G, pos, 
                                   node_size = node_size, 
                                   node_color = node_color_map, 
                                   ax = ax)
    nodes.set_edgecolor("xkcd:black")
    nodes.set_linewidth(1.5)
    nx.draw_networkx_labels(bp_complex_i_G, pos, labels = labels, 
                            font_size = node_font_size, font_weight = 'bold', 
                            font_color = 'xkcd:white', ax = ax)

    # return node position layout
    return pos


def display_PDB_network_for_complex(ax, chain_to_UniProt_mapping_dict, 
    interacting_UniProt_IDs, node_size, edge_width, node_font_size=10):
    '''
    Display network of interacting chains.
    
    This function displays a complete network in which nodes represent the 
    proteins in a specified PDB structure, and edges represent chains in that
    structure, using NetworkX. Edges that are classified as interacting
    (are < dist_threshold angstroms apart) are colored black.

    Parameters
    ----------
    ax object to draw on: Matplotlib Axes
    Mapping of Chains to UniProt IDs: dictionary
    List of Interacting Chains: list
    Size of Nodes in Network: int
    Width of Edges in Network: float
    Size of font for Node Labels: int (optional)

    Returns
    -------
    Node Positions
        Dictionary of Node Positions in NetworkX layout
    Interacting Network Edges
        List of Edges for Interacting Nodes
    Number of Network Edges
        Float of the Number of Possible Interacting Edges
    
    # what order are they returned?

    Examples
    --------
    # (1) Obtain list of interacting chains from 6YW7 structure
    # (2) Obtain a mapping of PDB ID 6YW7 chains to UniProt IDs
    # (3) Obtain list of interacting chains from 6YW7 
    # structure using UniProt IDs
    # (4) create figure and axis objects to draw on
    # (5) Visualize interacting chains using Uniprot IDs

    >>> interacting_chains_list = get_interacting_chains_from_PDB('6YW7', '.', 6)
    Downloading PDB structure '6YW7'...
    >>> chain_to_UniProt_mapping_dict = list_uniprot_pdb_mappings('6YW7')
    >>> interacting_UniProt_IDs = PDB_chains_to_uniprot(interacting_chains_list,chain_to_UniProt_mapping_dict)
    >>> fig, ax = plt.subplots()
    >>> node_layout_pdb, edges_list_pdb, num_possible_edges_pdb = display_PDB_network_for_complex(ax,chain_to_UniProt_mapping_dict, interacting_UniProt_IDs, 1500, 3, node_font_size = 8)
    >>> len(node_layout_pdb)
    7
    >>> len(edges_list_pdb)
    9
    >>> num_possible_edges_pdb
    21.0
    '''
    # create connected graph from all uniprot IDs
    chain_uniprot_IDs = ([chain_uniprot_i[0] for chain_uniprot_i in 
                          chain_to_UniProt_mapping_dict.values()])

    # create a graph from the nodes/genes of this complex structure
    pdb_structure_i_G = nx.Graph()
    pdb_structure_i_G.add_nodes_from(chain_uniprot_IDs)

    # iterate over physical interactions and add edges
    for chain_pair_interacting in interacting_UniProt_IDs:
        pdb_structure_i_G.add_edge(chain_pair_interacting[0], 
                                   chain_pair_interacting[1])

    # create a complete graph from the nodes of complex graph to add in all 
    # "background" edges (edges detected with AP-MS will be colored over)
    # position will be the same for both graphs since nodes are the same
    pdb_structure_i_G_complete = nx.Graph()
    pdb_structure_i_G_complete.add_nodes_from(pdb_structure_i_G.nodes)
    pdb_structure_i_G_complete.add_edges_from(
        itertools.combinations(pdb_structure_i_G.nodes, 2))

    # set position of nodes w/ circular layout
    pos = nx.circular_layout(pdb_structure_i_G)

    # construct edges for COMPLETE graph for "background" edges
    edges_complete = nx.draw_networkx_edges(pdb_structure_i_G_complete, 
                                pos, width = edge_width, alpha = 0.25, ax = ax)
    edges_complete.set_edgecolor("xkcd:grey")

    # construct edges
    edges = nx.draw_networkx_edges(pdb_structure_i_G, pos, 
                                   width = edge_width, ax = ax)
    edges.set_edgecolor('xkcd:blue')

    # construct nodes
    nodes = nx.draw_networkx_nodes(pdb_structure_i_G, pos, 
                                   node_size = node_size, 
                                   node_color = 'xkcd:black', ax = ax)
    nodes.set_edgecolor("xkcd:black")
    nodes.set_linewidth(1.5)
    nx.draw_networkx_labels(pdb_structure_i_G, pos, font_size = node_font_size,
                            font_weight = 'bold', font_color = 'xkcd:white', 
                            ax = ax)

    # return node position layout, list of edges detected, 
    # and number of possible edges
    return [pos, pdb_structure_i_G.edges, 
            float(len(pdb_structure_i_G_complete.edges))]

def display_PPI_network_match_PDB(ax, chain_to_UniProt_mapping_dict, 
    interacting_UniProt_IDs, bp_PPI_df, node_pos, node_size, edge_width, 
    node_font_size=10, bait_node_color='xkcd:red', 
    prey_node_color='xkcd:rose pink', AP_MS_edge_color='xkcd:red'):
    '''
    Display network of BioPlex PPIs for a set of interacting UniProt IDs.
    
    This function displays a complete network in which nodes represent the 
    proteins in a specified PDB structure, and edges represent chains in that
    structure, using NetworkX. Edges that are classified as interacting from
    BioPlex PPI data (detected through AP-MS) are colored darker.

    Parameters
    ----------
    ax object to draw on: Matplotlib Axes
    Mapping of Chains to UniProt IDs: dictionary
    List of Interacting Chains: list
    DataFrame of PPIs : Pandas DataFrame
    Networkx Position of Nodes: dict
    Size of Nodes in Network: int
    Width of Edges in Network: float
    Size of font for Node Labels: int (optional)
    Color of Nodes targeted as baits: str (optional)
    Color of Nodes detected as preys only: str (optional)
    Color of Edges observed via AP-MS from PPI data: str (optional)

    Returns
    -------
    Interacting Network Edges
        List of Edges for Interacting Nodes
    Number of Network Edges
        Float of the Number of Possible Interacting Edges
    

    Examples
    --------
    # (1) Obtain list of interacting chains from 6YW7 structure
    # (2) Obtain a mapping of PDB ID 6YW7 chains to UniProt IDs
    # (3) Obtain list of interacting chains from 6YW7 
    #     structure using UniProt IDs
    # (4) create figure and axis objects to draw on
    # (5) Visualize interacting chains using Uniprot IDs
    # (6) Get BioPlex PPI data
    # (7) create figure and axis objects to draw on
    # (8) Visualize BioPlex PPI interactions using layout 
    #     from interacting chains
    >>> interacting_chains_list = get_interacting_chains_from_PDB('6YW7', '.', 6)
    Downloading PDB structure '6YW7'...
    >>> chain_to_UniProt_mapping_dict = list_uniprot_pdb_mappings('6YW7')
    >>> interacting_UniProt_IDs = PDB_chains_to_uniprot(interacting_chains_list,chain_to_UniProt_mapping_dict)
    >>> fig, ax1 = plt.subplots()
    >>> node_layout_pdb, edges_list_pdb, num_possible_edges_pdb = display_PDB_network_for_complex(ax1, chain_to_UniProt_mapping_dict, interacting_UniProt_IDs, 1500, 3, node_font_size = 8)
    >>> bp_PPI_df = getBioPlex('293T', '3.0')
    >>> fig, ax2 = plt.subplots()
    >>> edges_list_bp, num_possible_edges_bp = display_PPI_network_match_PDB(ax2, chain_to_UniProt_mapping_dict, interacting_UniProt_IDs, bp_PPI_df, node_layout_pdb, 1500, 3, node_font_size = 8)
    >>> len(edges_list_bp)
    13
    >>> num_possible_edges_bp
    21.0
    '''
    # create connected graph from all uniprot IDs
    chain_uniprot_IDs = ([chain_uniprot_i[0] for chain_uniprot_i 
                          in chain_to_UniProt_mapping_dict.values()])
    
    # filter BioPlex PPI dataframe to include only interactions 
    # where both genes are found in complex
    structure_uniprots_i_PPI_filter = []
    for uniprot_A, uniprot_B in zip(bp_PPI_df.UniprotA, bp_PPI_df.UniprotB):

        # check for isoform IDs and adjust
        if '-' in uniprot_A:
            uniprot_A = uniprot_A.split('-')[0]
        if '-' in uniprot_B:
            uniprot_B = uniprot_B.split('-')[0]

        # check to see if both gene symbols for this 
        # interaction are genes in complex
        if ((uniprot_A in chain_uniprot_IDs) and 
            (uniprot_B in chain_uniprot_IDs)):
            structure_uniprots_i_PPI_filter.append(True)
        else:
            structure_uniprots_i_PPI_filter.append(False)

    structure_uniprots_i_PPI_filter = np.array(structure_uniprots_i_PPI_filter)
    # use filter to subset bp PPI dataframe
    bp_structure_i_df = bp_PPI_df[structure_uniprots_i_PPI_filter]
    bp_structure_i_df.reset_index(inplace = True, drop = True) # reset index

    # reconstruct UniprotA/UniprotB columns without '-' isoform id
    UniprotA_new = []
    UniprotB_new = []
    for UniprotA, UniprotB in zip(bp_structure_i_df.UniprotA, 
                                  bp_structure_i_df.UniprotB):
        if '-' in UniprotA:
            UniprotA_new.append(UniprotA.split('-')[0])
        else:
            UniprotA_new.append(UniprotA)

        if '-' in UniprotB:
            UniprotB_new.append(UniprotB.split('-')[0])
        else:
            UniprotB_new.append(UniprotB)

    # update columns for Uniprot source & 
    # Uniprot target to exclude isoform '-' ID
    bp_structure_i_df.loc[:,'UniprotA'] = UniprotA_new
    bp_structure_i_df.loc[:,'UniprotB'] = UniprotB_new

    # subset PPI dataframe to the cols we need to construct graph
    bp_structure_i_df = (bp_structure_i_df.loc[:,
                        ['UniprotA','UniprotB','SymbolA','SymbolB']])

    # create a graph from the nodes/genes of specified complex
    bp_structure_i_G = nx.Graph()
    bp_structure_i_G.add_nodes_from(chain_uniprot_IDs)

    # iterate over AP-MS interactions in PPI df and add edges
    for source, target in zip(bp_structure_i_df.UniprotA, 
                              bp_structure_i_df.UniprotB):
        bp_structure_i_G.add_edge(source, target)

    # get a list of genes that were identifed as "baits" 
    # and "preys" for coloring nodes
    bp_structure_i_baits = list(set(bp_structure_i_df.UniprotA))
    bp_structure_i_preys = list(set(bp_structure_i_df.UniprotB))

    # color nodes according to whether they were present among "baits"
    # & "preys", just "baits", just "preys" or not detected 
    # in PPI data for this structure
    node_color_map = []
    for node_i_uniprot in bp_structure_i_G.nodes:

        # gene is present among baits & preys for the PPIs 
        # detected in this structure
        if ((node_i_uniprot in bp_structure_i_baits) and 
            (node_i_uniprot in bp_structure_i_preys)):
            node_color_map.append(bait_node_color)

        # gene is present among baits but NOT preys for the PPIs 
        # detected in this structure
        elif ((node_i_uniprot in bp_structure_i_baits) and 
              (node_i_uniprot not in bp_structure_i_preys)):
            node_color_map.append(bait_node_color)

        # gene is NOT present among baits but is present among preys for 
        # the PPIs detected in this structure
        elif ((node_i_uniprot not in bp_structure_i_baits) and 
              (node_i_uniprot in bp_structure_i_preys)):
            node_color_map.append(prey_node_color)

        # gene is NOT present among baits and is NOT present among preys for 
        # the PPIs detected in this complex
        elif ((node_i_uniprot not in bp_structure_i_baits) and 
              (node_i_uniprot not in bp_structure_i_preys)):
            node_color_map.append('0.7')
            
    # can use the complete graph from the pdb direct 
    # interaction graph since they have the same nodes
    # create a complete graph from the nodes of complex graph to 
    # add in all "background" edges (edges 
    # detected with AP-MS will be colored over)
    # position will be the same for both graphs since nodes are the same
    bp_structure_i_G_complete = nx.Graph()
    bp_structure_i_G_complete.add_nodes_from(bp_structure_i_G.nodes)
    bp_structure_i_G_complete.add_edges_from(
        itertools.combinations(bp_structure_i_G.nodes, 2))

    # construct edges for COMPLETE graph for "background" edges
    edges_complete = nx.draw_networkx_edges(bp_structure_i_G_complete, 
                                            node_pos, width = edge_width,
                                            alpha = 0.25, ax = ax)
    edges_complete.set_edgecolor("xkcd:grey")

    # construct edges
    edges = nx.draw_networkx_edges(bp_structure_i_G, node_pos, 
                                   width = edge_width, ax = ax)
    edges.set_edgecolor(AP_MS_edge_color)

    # construct nodes
    nodes = nx.draw_networkx_nodes(bp_structure_i_G, node_pos, 
                                   node_size = node_size, 
                                   node_color = node_color_map, ax = ax)
    nodes.set_edgecolor("xkcd:black")
    nodes.set_linewidth(1.5)
    nx.draw_networkx_labels(bp_structure_i_G, node_pos, 
                            font_size = node_font_size, 
                            font_weight = 'bold', font_color = 'xkcd:white', 
                            ax = ax)

    # return node position layout, list of edges detected,
    # and number of possible edges
    return [bp_structure_i_G.edges, float(len(bp_structure_i_G_complete.edges))]    


#!/usr/bin/env python

import io
import itertools

import anndata as ad
import pandas as pd
import requests
from pypdb import *


def getBioPlex(cell_line, version):
    '''
    Load BioPlex interactions data.
    
    This function loads BioPlex PPI data for
    cell lines HEK293T and HCT116, note we
    only have version 1.0 for HCT116 cells.
    
    Parameters
    ----------
    cell_line : str
        Takes input ['293T','HCT116'].
    version : str
        Takes input ['3.0','1.0','2.0'].
    
    Returns
    -------
    Pandas DataFrame
        A dataframe with each row corresponding to a PPI interaction.
    
    Examples
    --------
    >>> bp_293t = getBioPlex('293T', '1.0')
    >>> bp_293t.head(1)
       GeneA   GeneB UniprotA UniprotB SymbolA SymbolB            pW       pNI      pInt
    0    100  728378   P00813   A5A3E0     ADA   POTEF  2.605947e-09  0.000333  0.999667
    '''
    if (f'{cell_line}.{version}' not in 
        ['293T.1.0','293T.2.0','293T.3.0','HCT116.1.0']):
        print('dataset not available for this Cell Line - Version')
        
    else:
        if f'{cell_line}.{version}' == '293T.1.0':
            file_ext = 'interactionList_v2'
        elif f'{cell_line}.{version}' == '293T.2.0':
            file_ext = 'interactionList_v4a'
        elif f'{cell_line}.{version}' == '293T.3.0':
            file_ext = '293T_Network_10K_Dec_2019'
        elif f'{cell_line}.{version}' == 'HCT116.1.0':
            file_ext = 'HCT116_Network_5.5K_Dec_2019'

        BioPlex_interactions_df = pd.read_csv(
                f"https://bioplex.hms.harvard.edu/data/BioPlex_{file_ext}.tsv", 
                sep = '\t')
        
    # if pulling 293T cell line version 1.0 or 2.0, change column names to 
    # standardize across datasets for input into other functions
    if (cell_line == '293T') and (version == '1.0'):
        BioPlex_interactions_df.rename({'Gene A':'GeneA','Gene B':'GeneB',
            'Uniprot A':'UniprotA','Uniprot B':'UniprotB',
            'Symbol A':'SymbolA','Symbol B':'SymbolB','p(Wrong)':'pW',
            'p(No Interaction)':'pNI','p(Interaction)':'pInt'}, 
            axis = 1, inplace = True)

    if (cell_line == '293T') and (version == '2.0'):
        BioPlex_interactions_df.rename({'p(Wrong)':'pW',
            'p(No Interaction)':'pNI','p(Interaction)':'pInt'}, 
            axis = 1, inplace = True)
    
    return BioPlex_interactions_df

def getGSE122425():
    '''
    Retrieve HEK293 RNAseq expression data.
    
    Returns
    -------
    adata : AnnData object
        SummarizedExperiment of HEK293 raw count with an 
        added layer storing rpkm.
    
    Examples
    --------
    >>> HEK293_adata = getGSE122425()
    >>> HEK293_adata
    AnnData object with n_obs × n_vars = 57905 × 6
        obs: 'SYMBOL', 'KO', 'GO', 'length'
        layers: 'rpkm'
    >>> print(HEK293_adata.obs_names[:10].tolist())
    ['ENSG00000223972', 'ENSG00000227232', 'ENSG00000243485', 'ENSG00000237613', 'ENSG00000268020', 'ENSG00000240361', 'ENSG00000186092', 'ENSG00000238009', 'ENSG00000239945', 'ENSG00000233750']
    '''
    # specify URL where data is stored
    baseURL = ('https://ftp.ncbi.nlm.nih.gov/geo/series/GSE122nnn/'
               'GSE122425/suppl/')
    filename = 'GSE122425_all.counts.293_vs_293NK.edgeR_all.xls.gz'
    outFilePath = filename[:-3]

    # stream the file as bytes into memory using 
    # io.bytesIO and decompress using pandas
    response = requests.get(baseURL + filename)
    content = response.content
    GSE122425_df = pd.read_csv(io.BytesIO(content), 
                               sep='\t', compression='gzip')

    # create annot for observations (rows)
    obs = GSE122425_df.loc[:,['gene_id','GeneSymbol','KO','GO','length']]
    obs.set_index('gene_id',inplace=True)
    obs.rename(mapper={'GeneSymbol':'SYMBOL'}, inplace=True, axis=1)

    # we have raw counts and rpkms here in one matrix
    # raw counts
    raw_X = (GSE122425_df.loc[:,
                              ['HEK293NK-SEQ1','HEK293NK-SEQ2',
                               'HEK293NK-SEQ3','HEK293-SEQ1','HEK293-SEQ2',
                               'HEK293-SEQ3']].values)
    # annot for variables (cols)
    raw_var = pd.DataFrame(index=['NK.1','NK.2','NK.3','WT.1','WT.2','WT.3'])
    
    # convert to AnnData object (default datatype is 'float32')
    adata = ad.AnnData(raw_X, obs=obs, var=raw_var, dtype='int32')

    # store rpkms as a layer
    rpkm_X = (GSE122425_df.loc[:,['HEK293NK-SEQ1_RPKM','HEK293NK-SEQ2_RPKM',
                                  'HEK293NK-SEQ3_RPKM','HEK293-SEQ1_RPKM',
                                  'HEK293-SEQ2_RPKM',
                                  'HEK293-SEQ3_RPKM']].values)
    adata.layers["rpkm"] = rpkm_X

    return adata

def getCorum(complex_set = 'all', organism = 'Human'):
    '''
    Functionality for retrieving the CORUM protein complex data.

    Parameters
    ----------
    complex_set : str
        Takes input ['all','drug','splice','partial'] (default 'all').
        Maps to CORUM files: 'all' -> corum_allComplexes.txt, 
        'drug' -> corum_drugs.txt, 
        'splice' -> corum_spliceComplexes.txt,
        'partial' -> corum_partialComplexes.txt.
    organism : str
        Takes input ['Bovine','Dog','Hamster','Human','MINK','Mammalia',
        'Mouse','Pig','Rabbit','Rat'] (default 'Human').

    Returns
    -------
    Pandas DataFrame
        A dataframe with each row corresponding to a CORUM complex.

    Examples
    --------
    >>> CORUM_df = getCorum()
    >>> CORUM_df = getCorum('splice', 'Human')
    >>> CORUM_df.size > 0
    True
    >>> 'complex_id' in CORUM_df.columns
    True
    >>> 'subunits_uniprot_id' in CORUM_df.columns
    True
    '''
    # Map complex_set input to CORUM filename
    complex_set_map = {
        'all': 'corum_allComplexes.txt',
        'partial': 'corum_partialComplexes.txt',
        'drug': 'corum_drugs.txt',
        'splice': 'corum_spliceComplexes.txt'
    }
    
    if complex_set not in complex_set_map:
        raise ValueError(f"complex_set must be one of {list(complex_set_map.keys())}, "
                        f"got '{complex_set}'")
    
    # specify URL
    baseURL = 'https://zenodo.org/records/17419058/files/'
    filename = complex_set_map[complex_set]

    # stream the file as bytes into memory using
    # io.BytesIO and read using pandas
    response = requests.get(baseURL + filename)
    response.raise_for_status()  # raise exception for bad status codes
    content = response.content
    CORUM_df = pd.read_csv(io.BytesIO(content), sep='\t')

    # filter to keep only CORUM sets for a specific organism
    CORUM_df = CORUM_df[CORUM_df.organism == organism]
    CORUM_df.reset_index(inplace = True, drop = True)
    
    return CORUM_df

def get_UniProts_from_CORUM(Corum_DF, Complex_ID):
    '''
    Retrieve set of UniProt IDs corresponding to a CORUM complex ID.
    
    This function takes a CORUM complex ID and CORUM complex DataFrame
    and returns the corresponding UniProt IDs.

    Parameters
    ----------
    DataFrame of CORUM complexes : Pandas DataFrame
    Corum Complex ID: int

    Returns
    -------
    UniProt IDs
        A list of UniProt IDs for the CORUM complex specified.

    Examples
    --------
    # (1) Obtain CORUM complexes
    # (2) Get set of UniProt IDs for specified protein 
    #     complex (Arp 2/3 complex ID: 27)
    >>> Corum_DF = getCorum('all', 'Human')
    >>> UniProts_Arp_2_3 = get_UniProts_from_CORUM(Corum_DF, Complex_ID = 27)
    >>> len(UniProts_Arp_2_3) > 0
    True
    >>> isinstance(UniProts_Arp_2_3, list)
    True
    '''
    # get UniProt IDs for each protein in the CORUM complex
        # NOTE: CORUM 5.1 now has a uniprot mapping txt file, not sure if better than DF?
    uniprot_IDs_list = (Corum_DF[Corum_DF.complex_id == Complex_ID].loc[:,
                                'subunits_uniprot_id'].values[0].split(';'))
    return uniprot_IDs_list

def get_PDB_from_UniProts(uniprot_IDs_list):
    '''
    Retreive PDB IDs for protein structures corresponding to set of UniProt IDs.
    
    This function takes a list of UniProt IDs and maps the corresponding 
    UniProt IDs (from the UniProt IDs input or CORUM complex ID) to PDB IDs 
    using the SIFTS project. Some metadata for each PDB ID is pulled from PDB 
    and stored in a DataFrame that is returned.

    Parameters
    ----------
    UniProt IDs : list

    Returns
    -------
    PDB IDs and associated metadata
        Pandas DataFrame of PDB IDs that map to the UniProt IDs input, 
        or corresponding UniProt IDs from the CORUM complex specified.

    Examples
    --------

    # this test is not consistent > PDB_ID_Arp_2_3.UniProts_mapped_to_PDB[1]
    ['O15144', 'P61158', 'P61160', 'O15145', 'P59998']
    # (1) Get set of PDB IDs for list of UniProt IDs that correspond to Arp 2/3
    >>> PDB_ID_Arp_2_3 = get_PDB_from_UniProts(['Q92747','O15144','P61158','P61160','O15145','P59998','O15511'])
    >>> PDB_ID_Arp_2_3.size
    15
    >>> type(PDB_ID_Arp_2_3)
    <class 'pandas.core.frame.DataFrame'>
    '''
    # get number of proteins in query
    num_proteins = len(uniprot_IDs_list)

    # Map from CORUM complex subunits given as UniProt IDs 
    # via [SIFTS](https://www.ebi.ac.uk/pdbe/docs/sifts/quick.html) 
    # to PDB structures:
    # "A summary of the UniProt to PDB mappings showing the UniProt accession 
    #  followed by a semicolon-separated list of PDB four letter codes."
    uniprot_pdb_mapping_df = pd.read_csv(
            ('ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/'
            'flatfiles/csv/uniprot_pdb.csv.gz'), header = 1, 
             sep = ',', compression = 'gzip')

    # set UniProt IDs as index
    uniprot_pdb_mapping_df.set_index('SP_PRIMARY', drop = True, inplace = True)

    # convert col PDB semicolon-separated list into Python list
    uniprot_pdb_mapping_df.loc[:,'PDB'] = ([PDB_codes_i.split(';') 
                                for PDB_codes_i in uniprot_pdb_mapping_df.PDB])

    # get PDB IDs that map to each UniProt ID
    PDB_IDs_for_uniprot_dict = {}
    for uniprot_ID_i in uniprot_IDs_list:

        # check to see if UniProt ID exists in mapping
        if uniprot_ID_i in uniprot_pdb_mapping_df.index:

            # append to list of PDB IDs, take ALL PDB IDs in mapped list
            mapped_PDB_ID_i = (uniprot_pdb_mapping_df.loc[uniprot_ID_i,
                                                          :].values[0])

            # convert to uppercase
            mapped_PDB_ID_i = [PDB_ID.upper() for PDB_ID in mapped_PDB_ID_i]

            PDB_IDs_for_uniprot_dict[uniprot_ID_i] = mapped_PDB_ID_i

        else:
            print(f'WARNING: {uniprot_ID_i} does not have any '
                  'corresponding PDB IDs mapped.')

    # create dictionary of PDB IDs and store list of 
    # Uniprot IDs each one is mapped to
    # flatten list of PDB IDs that mapped to UniProt IDs
    unique_PDB_IDs = list(set(
            itertools.chain(*list(PDB_IDs_for_uniprot_dict.values()))))
    uniprot_IDs_list_for_PDB_dict = {}
    for PDB_ID in unique_PDB_IDs:

        uniprot_to_PDB_i = []
        for uniprot_ID in PDB_IDs_for_uniprot_dict.keys():
            if PDB_ID in PDB_IDs_for_uniprot_dict[uniprot_ID]:
                uniprot_to_PDB_i.append(uniprot_ID)

        uniprot_IDs_list_for_PDB_dict[PDB_ID] = uniprot_to_PDB_i

    uniprot_IDs_list_for_PDB_series = pd.Series(uniprot_IDs_list_for_PDB_dict)

    # if no PDB IDs mapped to UniProt IDs (empty list), raise warning
    if len(uniprot_IDs_list_for_PDB_series) == 0:
        print('WARNING: Could not map PDB ID to this CORUM '
              'complex ID or UniProt IDs.')
        complex_i_PDBs_df = None

    else:
        # iterate through PDB ID & retreive metadata
        # stores number of polymer proteins for this structure
        PDB_protein_count = []
        # store the date of deposit for this structure
        PDB_deposit_date = []
        # store the title of the citation for this structure
        PDB_citation_title = []
        # store list of UniProt IDs that mapped to each PDB ID
        PDB_to_uniprot_map_list = []
        for PDB_ID in uniprot_IDs_list_for_PDB_series.index:

            # retreive metadata for this structure from PDB
            PDB_structure_all_info = get_info(PDB_ID)
            PDB_protein_count.append(
                (PDB_structure_all_info['rcsb_entry_info']
                ['polymer_entity_count_protein']))
            PDB_deposit_date.append(
                (pd.to_datetime(PDB_structure_all_info['rcsb_accession_info']
                ['deposit_date'])))
            PDB_citation_title.append(
                PDB_structure_all_info['rcsb_primary_citation']['title'])
            PDB_to_uniprot_map_list.append(
                uniprot_IDs_list_for_PDB_series[PDB_ID])

        # convert CORUM complex i - associated PDB IDs into 
        # DataFrame w/ # proteins & resolution
        UniProt_assoc_PDBs_df = (pd.DataFrame(
            index = uniprot_IDs_list_for_PDB_series.index))
        UniProt_assoc_PDBs_df.loc[:,'num_proteins'] = PDB_protein_count
        UniProt_assoc_PDBs_df.loc[:,'deposit_date'] = PDB_deposit_date
        UniProt_assoc_PDBs_df.loc[:,'citation_title'] = PDB_citation_title
        # the UniProt IDs that mapped to this PDB ID from SIFTS
        (UniProt_assoc_PDBs_df.loc[:,
            'UniProts_mapped_to_PDB']) = PDB_to_uniprot_map_list

        # column for number of proteins in PDB structure different from
        # num proteins listed in CORUM complex or input by user
        (UniProt_assoc_PDBs_df.loc[:,
            'num_proteins_diff_btwn_PDB_and_UniProts_input']) = abs(
                UniProt_assoc_PDBs_df.num_proteins - num_proteins)

        # pick PDB structure that has same number of proteins/chains as CORUM
        # complex (or matches closest), then rank by most recent deposit date
        # tyl: recent deposit date fluctuates?
        UniProt_assoc_PDBs_df.sort_values(
            by = (['num_proteins_diff_btwn_PDB_and_UniProts_input',
                   'deposit_date']), 
            ascending = [True, False], inplace = True)
        
    return UniProt_assoc_PDBs_df


from .analysis_funcs import (
    PDB_chains_to_uniprot,
    bioplex2graph,
    get_DataFrame_from_PPI_network,
    get_interacting_chains_from_PDB,
    get_mappings_data,
    get_PPI_network_for_complex,
    get_prop_edges_in_complex_identified,
    list_uniprot_pdb_mappings,
    make_request,
    resampling_test_for_uniprot_list,
)
from .data_import_funcs import (
    get_PDB_from_UniProts,
    get_UniProts_from_CORUM,
    getBioPlex,
    getCorum,
    getGSE122425,
)
from .version import __version__
from .visualization_funcs import (
    display_PDB_network_for_complex,
    display_PPI_network_for_complex,
    display_PPI_network_match_PDB,
)




#!/usr/bin/env python

import itertools
import random
import re
from collections import Counter

import networkx as nx
import numpy as np
import pandas as pd
import requests
from Bio.PDB import *
from scipy.spatial.distance import cdist


def bioplex2graph(bp_PPI_df):
    '''
    Convert BioPlex PPIs into a graph.
    
    This function converts representation of BioPlex PPIs into a 
    graph data structure representation of BioPlex PPIs in a NetworkX 
    object from NetworkX. 

    Parameters
    ----------
    DataFrame of PPIs : Pandas DataFrame

    Returns
    -------
    NetworkX graph
        A NetworkX graph with Nodes = Uniprot Gene Symbols 
        and Edges = interactions.

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    # (2) Turn the data into a graph
    >>> bp_293t_df = getBioPlex('293T', '3.0')
    >>> bp_293t_G = bioplex2graph(bp_293t_df) 
    >>> type(bp_293t_G)
    <class 'networkx.classes.digraph.DiGraph'>
    >>> len(bp_293t_G.edges)
    115868
    >>> len(bp_293t_G.nodes)
    13689
    '''
    # add isoform columns for Uniprot source & Uniprot target
    bp_PPI_df.loc[:,'isoformA'] = bp_PPI_df.UniprotA
    bp_PPI_df.loc[:,'isoformB'] = bp_PPI_df.UniprotB

    # reconstruct UniprotA/UniprotB columns without '-' isoform id
    UniprotA_new = []
    UniprotB_new = []
    for UniprotA, UniprotB in zip(bp_PPI_df.UniprotA, bp_PPI_df.UniprotB):

        if '-' in UniprotA:
            UniprotA_new.append(UniprotA.split('-')[0])
        else:
            UniprotA_new.append(UniprotA)

        if '-' in UniprotB:
            UniprotB_new.append(UniprotB.split('-')[0])
        else:
            UniprotB_new.append(UniprotB)

    # update columns for Uniprot source 
    # & Uniprot target to exclude isoform '-' ID
    bp_PPI_df.loc[:,'UniprotA'] = UniprotA_new
    bp_PPI_df.loc[:,'UniprotB'] = UniprotB_new
    
    # construct graph from BioPlex PPI data
    bp_G = nx.DiGraph()
    for source, target, pW, pNI, pInt in zip(bp_PPI_df.UniprotA, 
            bp_PPI_df.UniprotB, bp_PPI_df.pW, bp_PPI_df.pNI, bp_PPI_df.pInt):
        bp_G.add_edge(source, target, pW=pW, pNI=pNI, pInt=pInt)
        
    # get mapping uniprot -> entrez & store as node attribute
    uniprot_entrez_dict = {}
    for uniprot_A, entrez_A in zip(bp_PPI_df.UniprotA, bp_PPI_df.GeneA):
        uniprot_entrez_dict[uniprot_A] = entrez_A
    for uniprot_B, entrez_B in zip(bp_PPI_df.UniprotB, bp_PPI_df.GeneB):
        uniprot_entrez_dict[uniprot_B] = entrez_B

    for node_i in bp_G.nodes():
        bp_G.nodes[node_i]["entrezid"] = uniprot_entrez_dict[node_i]

    # get mapping uniprot -> symbol & store as node attribute
    uniprot_symbol_dict = {}
    for uniprot_A, symbol_A in zip(bp_PPI_df.UniprotA, bp_PPI_df.SymbolA):
        uniprot_symbol_dict[uniprot_A] = symbol_A
    for uniprot_B, symbol_B in zip(bp_PPI_df.UniprotB, bp_PPI_df.SymbolB):
        uniprot_symbol_dict[uniprot_B] = symbol_B

    for node_i in bp_G.nodes():
        bp_G.nodes[node_i]["symbol"] = uniprot_symbol_dict[node_i]

    # get mapping uniprot -> isoform & store as node attribute
    uniprot_isoform_dict = {}
    for uniprot_A, isoform_A in zip(bp_PPI_df.UniprotA, bp_PPI_df.isoformA):
        uniprot_isoform_dict[uniprot_A] = isoform_A
    for uniprot_B, isoform_B in zip(bp_PPI_df.UniprotB, bp_PPI_df.isoformB):
        uniprot_isoform_dict[uniprot_B] = isoform_B

    for node_i in bp_G.nodes():
        bp_G.nodes[node_i]["isoform"] = uniprot_isoform_dict[node_i]
        
    # get set of baits & store a bait boolean as node attribute 
    # if node is a bait True & False otherwise
    bp_i_baits = set(bp_PPI_df.UniprotA)
    for node_i in bp_G.nodes():
        if node_i in bp_i_baits:
            bp_G.nodes[node_i]["bait"] = True
        else:
            bp_G.nodes[node_i]["bait"] = False
    
    return bp_G

def get_PPI_network_for_complex(bp_PPI_G, Corum_DF, Complex_ID):
    '''
    Retrieve Network of BioPlex (AP-MS) PPIs for a CORUM complex.
    
    This function returns a subgraph of PPIs identified through AP-MS
    between the proteins in a specified CORUM complex.

    Parameters
    ----------
    Network of PPIs : NetworkX graph
    DataFrame of CORUM complexes : Pandas DataFrame
    Corum Complex ID: int

    Returns
    -------
    NetworkX Graph
        A subgraph induced by the proteins in a CORUM complex 
        from the BioPlex network used as input.

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    # (2) Obtain NetworkX graph representation of 293T PPI network
    # (3) Obtain CORUM complexes
    # (4) Get AP-MS interactions as subgraph for a specified protein complex using PPI data
    >>> bp_293t_df = getBioPlex('293T', '3.0')
    >>> bp_293t_G = bioplex2graph(bp_293t_df)
    >>> Corum_DF = getCorum()
    >>> ING2_bp_293t_G = get_PPI_network_for_complex(bp_293t_G, Corum_DF, 2851)
    >>> type(ING2_bp_293t_G)
    <class 'networkx.classes.digraph.DiGraph'>
    >>> len(ING2_bp_293t_G)
    12
    '''
    # store gene UNIPROT IDs that belong to this complex in a list
    genes_in_complex_i = (Corum_DF[Corum_DF.complex_id == Complex_ID].loc[:,
                                'subunits_uniprot_id'].values[0].split(';'))
    
    # get subgraph induced by the subset of nodes in this CORUM complex
    bp_complex_i_G = bp_PPI_G.subgraph(genes_in_complex_i)
    
    return bp_complex_i_G

def get_DataFrame_from_PPI_network(bp_PPI_G):
    '''
    Convert Network of BioPlex (AP-MS) PPIs into DataFrame of BioPlex 
    interaction Network.
    
    This function returns a DataFrame of PPIs (identified through AP-MS) 
    represented as a graph.

    Parameters
    ----------
    Network of PPIs : NetworkX graph

    Returns
    -------
    Pandas DataFrame
        A DataFrame of edges (AP-MS interactions) from a network.

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    # (2) Obtain NetworkX graph representation of 293T PPI network
    # (3) Obtain CORUM complexes
    # (4) Get AP-MS interactions as subgraph for a specified protein complex using PPI data row corresponding to an edge
    # (5) Convert ING2 AP-MS network into DataFrame w/ each 
    # order of rows in dataframe?
    >>> bp_293t_df = getBioPlex('293T', '3.0')
    >>> bp_293t_G = bioplex2graph(bp_293t_df)
    >>> Corum_DF = getCorum()
    >>> ING2_bp_293t_G = get_PPI_network_for_complex(bp_293t_G, Corum_DF, 2851)
    >>> ING2_bp_293t_df = get_DataFrame_from_PPI_network(ING2_bp_293t_G)
    >>> type(ING2_bp_293t_df)
    <class 'pandas.core.frame.DataFrame'>
    >>> ING2_bp_293t_df.shape
    (51, 7)
    >>> ING2_bp_293t_df.columns
    Index(['UniprotA', 'UniprotB', 'SymbolA', 'SymbolB', 'pW', 'pNI', 'pInt'], dtype='object')
    '''
    # get list of edges in network
    PPI_edge_list = list(bp_PPI_G.edges)

    # make node_A and node_B columns for both UNIPROT & SYMBOLS
    uniprotA_list = []
    uniprotB_list = []
    symbolA_list = []
    symbolB_list = []

    # make columns for calcs detected for each edge
    pW_list = []
    pNI_list = []
    pInt_list = []

    # iterate through each edge to store data for each row of DataFrame
    for edge_i in PPI_edge_list:

        nodeA, nodeB = edge_i

        # nodes are labeled with UNIPROT
        uniprotA_list.append(nodeA)
        uniprotB_list.append(nodeB)

        # get gene SYMBOL
        symbolA_list.append(bp_PPI_G.nodes[nodeA]['symbol'])
        symbolB_list.append(bp_PPI_G.nodes[nodeB]['symbol'])

        # get AP-MS calculations
        pW_list.append(bp_PPI_G.edges[edge_i]['pW'])
        pNI_list.append(bp_PPI_G.edges[edge_i]['pNI'])
        pInt_list.append(bp_PPI_G.edges[edge_i]['pInt'])

    # convert lists into cols of DataFrame
    bp_complex_i_df = pd.DataFrame()
    bp_complex_i_df.loc[:,'UniprotA'] = uniprotA_list
    bp_complex_i_df.loc[:,'UniprotB'] = uniprotB_list
    bp_complex_i_df.loc[:,'SymbolA'] = symbolA_list
    bp_complex_i_df.loc[:,'SymbolB'] = symbolB_list
    bp_complex_i_df.loc[:,'pW'] = pW_list
    bp_complex_i_df.loc[:,'pNI'] = pNI_list
    bp_complex_i_df.loc[:,'pInt'] = pInt_list
    
    return bp_complex_i_df
    
def get_prop_edges_in_complex_identified(bp_PPI_G, Corum_DF, Complex_ID):
    '''
    Calculates proportion of all possible edges identified from BioPlex (AP-MS) 
    PPIs for a CORUM complex.
    
    This function returns the proportion of all possible PPIs identified 
    through AP-MS between the proteins in a specified CORUM complex.

    Parameters
    ----------
    DataFrame of PPIs : Pandas DataFrame
    DataFrame of CORUM complexes : Pandas DataFrame
    Corum Complex ID: int

    Returns
    -------
    Float
        The proportion of interactions between all proteins in CORUM complex 
        identified through AP-MS PPI data

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    # (2) Obtain NetworkX graph representation of 293T PPI network
    # (3) Obtain CORUM complexes
    # (4) Get proportion of interactions identified for a specified CORUM complex 
    # using PPI data

    >>> bp_293t_df = getBioPlex('293T', '3.0')
    >>> bp_293t_G = bioplex2graph(bp_293t_df)
    >>> Corum_DF = getCorum()
    >>> get_prop_edges_in_complex_identified(bp_293t_G, Corum_DF, 2851)
    '''
    # store gene UNIPROT IDs that belong to this complex in a list
    genes_in_complex_i = (Corum_DF[Corum_DF.complex_id == Complex_ID].loc[:,
                                'subunits_uniprot_id'].values[0].split(';'))
    
    # get subgraph induced by the subset of nodes in this CORUM complex
    bp_complex_i_G = bp_PPI_G.subgraph(genes_in_complex_i)
        
    # create a complete graph from the nodes of complex graph 
    # (all possible interactions between proteins)
    bp_complex_i_G_complete = nx.Graph()
    bp_complex_i_G_complete.add_nodes_from(genes_in_complex_i)
    bp_complex_i_G_complete.add_edges_from(
                    itertools.combinations(genes_in_complex_i, 2))
    
    # calculate proportion of interactions between proteins in complex 
    # identified through AP-MS
    prop_edges_identified = (float(len(list(bp_complex_i_G.edges)))/
                             float(len(list(bp_complex_i_G_complete.edges))))
    
    # return proportion of edges ID'd through AP-MS, round to 3 decimal places
    return round(prop_edges_identified, 3)

def resampling_test_for_uniprot_list(bp_PPI_G, uniprot_list, 
                    num_resamples = 1000, preserve_node_degree = False):
    '''
    Run resampling test to check for enrichment of BioPlex (AP-MS) PPIs 
    for a given list of proteins (uniprot IDs).

    This function returns a p-value after running a resampling test by 
    1. taking the number of proteins in the specified list of uniprot IDs (N), 
    2. choosing N random proteins from the Graph generated by all of the 
       PPI data (G),
    3. calculating the number of edges in the Subgraph (S) induced by N random 
       proteins (with the same proportion of baits (+/- 10%) as the complex) 
       and storing this value (E_i),
    4. if preserve_node_degree option is invoked, then baits & preys in S 
       must have the same degree distribution as baits & preys in the
       network generated by given uniprot IDs, respectively,
    5. repeating steps 1-3 num_resamples times to create a null distribution, 
    6. calculating the number of edges between N proteins in the 
       complex (E), 
    7. returning a p-value by calculating the proportion of values 
       [E_1, E_2, ... , E_num_resamples] that are greater than or equal to E.

    Parameters
    ----------
    Network of PPIs : NetworkX graph
    list of uniprot IDs : list
    Number of resamplings: int
    option to preserve degree distribution in subgraphs : bool

    Returns
    -------
    Float
        A p-value from a resampling test to check for enrichment of PPIs 
        detected between proteins in list

    Examples
    --------
    # (1) Obtain the latest version of the 293T PPI network
    >>> bp_293t_df = getBioPlex('293T', '3.0')
    # (2) Obtain NetworkX graph representation of 293T PPI network
    >>> bp_293t_G = bioplex2graph(bp_293t_df)
    # (3) Obtain CORUM complexes
    >>> Corum_DF = getCorum()
    # (4) Get list of uniprots for Arp2/3 complex
    >>> UniProts_Arp_2_3 = get_UniProts_from_CORUM(Corum_DF, Complex_ID = 27)
    # (5) Calculate p-value to check for enrichment of edges in 
    #     Arp2/3 protein complex
    >>> resampling_test_for_CORUM_complex(bp_293t_G, UniProts_Arp_2_3, 1000)
    '''
    # get subgraph induced by the subset of nodes in this protein list
    bp_uniprots_i_G = bp_PPI_G.subgraph(uniprot_list)

    # number of edges detected between proteins in this CORUM 
    # complex among PPI data
    num_edges_identified_uniprot_list = float(len(list(bp_uniprots_i_G.edges)))

    # if no edges detected in this uniprot list, return an ERROR message
    if num_edges_identified_uniprot_list == 0.0:
        print('ERROR: no edges detected in PPI data for this protein list, '
              'p-value could not be computed.')

    # if at least 1 edge in complex, estimate p-value 
    # using resampling test
    else:

        # number of genes in complex (N genes)
        num_genes_in_uniprot_list = len(list(bp_uniprots_i_G.nodes))    

        # list of nodes in large network generated from PPI data
        nodes_in_overall_PPI_network = list(bp_PPI_G.nodes)
        
        # bait degrees in large PPI network
        ########################################
        # create a filter for uniprot IDs in large network that are baits
        G_baits_filter = np.array(
            [bp_PPI_G.nodes[node_i]['bait'] for node_i in bp_PPI_G.nodes])

        # get array of baits in large network
        G_baits = np.array(nodes_in_overall_PPI_network)[G_baits_filter]

        # get degree of each bait
        G_baits_degrees = bp_PPI_G.degree(G_baits)

        # prey degrees in large PPI network
        ########################################
        # create a filter for uniprot IDs in complex that are preys
        G_preys_filter = np.array(
            [not bp_PPI_G.nodes[node_i]['bait'] for node_i in bp_PPI_G.nodes])

        # get array of preys in large network
        G_preys = np.array(nodes_in_overall_PPI_network)[G_preys_filter]

        # get degree of each prey
        G_preys_degrees = bp_PPI_G.degree(G_preys)

        # bait degrees in complex
        ########################################
        # get array of uniprot IDs in complex
        bp_uniprots_i_G_nodes = np.array(bp_uniprots_i_G.nodes)

        # create a filter for uniprot IDs in complex that are baits
        bp_uniprots_i_G_baits_filter = np.array(
            [bp_uniprots_i_G.nodes[node_i]['bait'] for 
                 node_i in bp_uniprots_i_G.nodes])

        # get list of baits in complex
        bp_uniprots_i_G_baits = bp_uniprots_i_G_nodes[
            bp_uniprots_i_G_baits_filter]

        # get degree of each bait
        bp_uniprots_i_G_baits_degrees = bp_uniprots_i_G.degree(
                                        bp_uniprots_i_G_baits)

        # get degree distribution of baits
        bp_uniprots_i_G_baits_degree_distr = Counter(
            [bp_uniprots_i_G_baits_degrees[uniprot_i] for 
             uniprot_i in bp_uniprots_i_G_baits])
        
        # remove count for any "0" degree baits
        # would not contribute to edges in S
        if 0 in bp_uniprots_i_G_baits_degree_distr.keys():
            del bp_uniprots_i_G_baits_degree_distr[0]
        
        # find number of baits in complex
        
        # if preserve node degree TRUE
        # exclude baits w/ degree 0 from bait count
        # b/c random subgraphs won't have these
        if preserve_node_degree == True:
            bp_complex_i_baits_num = (np.sum(list(
                    bp_uniprots_i_G_baits_degree_distr.values())))
        
        # else count all baits in complex
        else:
            bp_complex_i_baits_num = (np.sum(bp_uniprots_i_G_baits_filter))
        
        # find proportion of baits in complex
        bp_complex_i_baits_prop = (float(bp_complex_i_baits_num) / 
                                   float(num_genes_in_uniprot_list))

        # prey degrees in complex
        ########################################
        # get array of uniprot IDs in complex
        bp_uniprots_i_G_nodes = np.array(bp_uniprots_i_G.nodes)

        # create a filter for uniprot IDs in complex that are preys
        bp_uniprots_i_G_preys_filter = np.array(
            [not bp_uniprots_i_G.nodes[node_i]['bait'] for 
             node_i in bp_uniprots_i_G.nodes])

        # get list of prey in complex
        bp_uniprots_i_G_preys = bp_uniprots_i_G_nodes[
            bp_uniprots_i_G_preys_filter]

        # get degree of each prey
        bp_uniprots_i_G_preys_degrees = bp_uniprots_i_G.degree(
            bp_uniprots_i_G_preys)

        # get degree distribution of preys
        bp_uniprots_i_G_preys_degree_distr = Counter(
            [bp_uniprots_i_G_preys_degrees[uniprot_i] for 
             uniprot_i in bp_uniprots_i_G_preys])
        
        # remove count for any "0" degree preys
        # would not contribute to edges in S
        if 0 in bp_uniprots_i_G_preys_degree_distr.keys():
            del bp_uniprots_i_G_preys_degree_distr[0]

        # list that will store number of edges detected in each subgraph
        num_edges_random_subgraphs = []

        # iterate through num_resamples random subgraphs induced by N nodes
        S_i = 0
        while S_i < num_resamples:
            
            # if degree distribution option is invoked
            if preserve_node_degree == True:

                # baits in S
                ########################################
                # preserve degrees by randomly pulling baits
                # with same degree for each uniprot in complex
                S_rando_baits = []
                for deg_i, uniprot_count_i in zip(
                    bp_uniprots_i_G_baits_degree_distr.keys(), 
                    bp_uniprots_i_G_baits_degree_distr.values()):

                    # get baits in G with same degree
                    G_baits_with_deg_i = list(
                        G_baits[np.array(
                            [G_baits_degrees[bait_i] == deg_i for 
                             bait_i in G_baits])])
                    
                    # choose N baits w/ same degree at random w/o replacement
                    N_rando_baits_from_PPI_network = random.sample(
                        G_baits_with_deg_i, uniprot_count_i)

                    # add to list of random baits
                    S_rando_baits = (S_rando_baits + 
                                     N_rando_baits_from_PPI_network)

                # preys in S
                ########################################
                # preserve degrees by randomly pulling preys
                # with same degree for each uniprot in complex
                S_rando_preys = []
                for deg_i, uniprot_count_i in zip(
                    bp_uniprots_i_G_preys_degree_distr.keys(), 
                    bp_uniprots_i_G_preys_degree_distr.values()):

                    # get preys in G with same degree
                    G_preys_with_deg_i = list(
                        G_preys[np.array([G_preys_degrees[prey_i] == deg_i for 
                                          prey_i in G_preys])])

                    # choose N preys w/ same degree at random w/o replacement
                    N_rando_preys_from_PPI_network = random.sample(
                        G_preys_with_deg_i, uniprot_count_i)

                    # add to list of random preys
                    S_rando_preys = (S_rando_preys + 
                                     N_rando_preys_from_PPI_network)

                # combine random Preys & Baits while preserving degrees
                N_rando_nodes_from_PPI_network = S_rando_baits + S_rando_preys

            # if degree distribution not invoked
            elif preserve_node_degree == False:

                # choose N genes at random without replacement
                N_rando_nodes_from_PPI_network = random.sample(
                    nodes_in_overall_PPI_network, num_genes_in_uniprot_list)

            # get subgraph induced by random subset of nodes
            bp_PPI_S = bp_PPI_G.subgraph(N_rando_nodes_from_PPI_network)

            # check to see if nodes in subgraph have the same proportion of 
            # baits as the subgraph induced by the CORUM complex
            # excluding baits that had degree 0
            bp_S_baits_num = np.sum(
                [bp_PPI_S.nodes[node_i]['bait'] for node_i in bp_PPI_S.nodes])
            bp_S_baits_prop = (float(bp_S_baits_num) / 
                               float(num_genes_in_uniprot_list))

            # proportion of baits in CORUM complex & S are the same (+/- 10%)
            if abs(bp_complex_i_baits_prop - bp_S_baits_prop) <= 0.1:

                # calculate the number of edges detected within 
                # subgraph induced by random nodes
                num_edges_S = float(len(list(bp_PPI_S.edges)))

                # store in list that contains resamplings
                num_edges_random_subgraphs.append(num_edges_S)

                # count this as a resampling
                S_i += 1

        # convert list to numpy array
        num_edges_random_subgraphs = np.array(num_edges_random_subgraphs)

        # calculate proportion of subgraphs that had more edges than edges 
        # detected in complex (p-val from resampling test)
        p_val = (float(np.sum(num_edges_random_subgraphs 
                                >= num_edges_identified_uniprot_list) + 1.0) / 
                            (float(num_resamples) + 1.0))
    return p_val

def get_interacting_chains_from_PDB(PDB_ID_structure_i, protein_structure_dir, dist_threshold):
    '''
    Retreive chain pairs that are physically close to eachother from 
    PDB structure.
    
    This function downloads the PDB structure that is specified from the input 
    PDB ID into the input directory, then computes the pairwise distances 
    between all atoms for each pair of chains in the structure. A list of 
    chain pairs that are interacting (have at least a pair of 
    atoms < dist_threshold angstroms apart) is returned.

    Parameters
    ----------
    PDB ID: str
    directory to store PDB file: str
    distance threshold: int

    Returns
    -------
    Interacting Chains
        List of chain pairs from PDB structure that have at least 
        one pair of atoms located < distance threshold apart.

    Examples
    --------
    # (1) Obtain list of interacting chains from 6YW7 structure
    >>> interacting_chains_list = get_interacting_chains_from_PDB('6YW7', '.', 6)
    Downloading PDB structure '6YW7'...
    >>> interacting_chains_list
    [['A', 'D'], ['A', 'E'], ['A', 'B'], ['D', 'F'], ['B', 'F'], ['B', 'G'], ['F', 'G'], ['F', 'C'], ['G', 'C']]
    '''
    # download structure from PDB
    pdbl = PDBList()
    PBD_file_path = pdbl.retrieve_pdb_file(PDB_ID_structure_i, 
                                           pdir=protein_structure_dir, 
                                           file_format='pdb', 
                                           overwrite=True)

    # create a structure object
    parser = PDBParser()
    structure = parser.get_structure(PDB_ID_structure_i, PBD_file_path)

    model = structure[0]
    chain_IDs = [chain.get_id() for chain in model] # get a list of all chains

    # we want to test every pair of chains to see if they have any atoms 
    # that are < 6 angstroms in distance
    possible_chain_pairs = list(itertools.combinations(chain_IDs, 2))

    chain_pairs_direct_interaction = []
    # iterate through all chain pairs and check to see if any atoms are close
    for chain_i_id, chain_j_id in possible_chain_pairs:

        # get chain objects from models
        chain_i = model[chain_i_id]
        chain_j = model[chain_j_id]

        # get all atoms from each chain, 'A' stands for ATOM
        atom_list_i = Selection.unfold_entities(chain_i, "A")
        atom_list_j = Selection.unfold_entities(chain_j, "A")

        # get the coordinates for the atom in each chain
        atom_coords_i = ([atom_list_i[k].coord 
                          for k in range(0,len(atom_list_i))])
        atom_coords_i = np.vstack(atom_coords_i)

        atom_coords_j = ([atom_list_j[k].coord 
                          for k in range(0,len(atom_list_j))])
        atom_coords_j = np.vstack(atom_coords_j)

        # compute pairwise distances betweeen all atoms from different chains
        dists = cdist(atom_coords_i, atom_coords_j)

        # if a pair of atoms < dist_threshold angstroms apart, store as interacting chains
        if np.sum(dists < dist_threshold) >= 1:
            chain_pairs_direct_interaction.append([chain_i_id, chain_j_id])
        
    return chain_pairs_direct_interaction

def make_request(url, mode, pdb_id):
    '''
    Make requests to PDBe API.
    
    This function can make GET and POST requests to the PDBe API.
    
    Parameters
    ----------
    url: str
    mode: str
    pdb_id: str
    
    Returns
    -------
        JSON or None
    '''
    if mode == "get":
        response = requests.get(url=url+pdb_id)
    elif mode == "post":
        response = requests.post(url, data=pdb_id)

    if response.status_code == 200:
        return response.json()
    else:
        print("[No data retrieved - %s] %s" 
              % (response.status_code, response.text))
    
    return None

def get_mappings_data(pdb_id):
    '''
    Get mappings data for PDB ID.
    
    This function will retreive the mappings data from
    the PDBe API using the make_request() function.
    
    Parameters
    ----------
    pdb_id: str
    
    Returns
    -------
        JSON of mappings or None
    '''
    # specify URL
    base_url = "https://www.ebi.ac.uk/pdbe/"
    api_base = base_url + "api/"
    uniprot_mapping_url = api_base + 'mappings/uniprot/'
    
    # Check if the provided PDB id is valid
    # There is no point in making an API call
    # with bad PDB ids
    if not re.match("[0-9][A-Za-z][A-Za-z0-9]{2}", pdb_id):
        print("Invalid PDB id")
        return None
    
    # GET the mappings data
    mappings_data = make_request(uniprot_mapping_url, "get", pdb_id)
    
    # Check if there is data
    if not mappings_data:
        print("No data found")
        return None
    
    return mappings_data

def list_uniprot_pdb_mappings(pdb_id):
    '''
    Get PDB chain to UniProt mappings.
    
    This function retrieves PDB > UniProt mappings using the 
    get_mappings_data() function, the parses the resulting 
    JSON to construct a dictionary where each key is a chain
    from the PDB structure, and the corresponding value for
    each is a list of UniProt IDs that map to the chain from 
    the SIFTS project.
    
    Parameters
    ----------
    pdb_id: str
    
    Returns
    -------
    Chain to UniProt Map
        Dictionary of PDB ID chain to UniProt ID mappings
        
    Examples
    --------
    # (1) Obtain a mapping of PDB ID 6YW7 chains to UniProt IDs
    >>> chain_to_UniProt_mapping_dict = list_uniprot_pdb_mappings('6YW7')
    >>> len(chain_to_UniProt_mapping_dict)
    7
    >>> sorted(chain_to_UniProt_mapping_dict.keys())
    ['A', 'B', 'C', 'D', 'E', 'F', 'G']
    >>> sorted(chain_to_UniProt_mapping_dict.values())
    [['O15144'], ['O15145'], ['O15511'], ['P59998'], ['P61158'], ['P61160'], ['Q92747']]
    '''
    # convert to PDB id to lower case
    pdb_id = pdb_id.lower()
    
    # Getting the mappings data
    mappings_data = get_mappings_data(pdb_id)
    
    # If there is no data, return None
    if not mappings_data:
        return None
    
    # dictionary that stores UniProt > chain id mappings
    uniprot_chain_mapping_dict = {}
    
    uniprot = mappings_data[pdb_id]["UniProt"]
    for uniprot_id in uniprot.keys():
        mappings = uniprot[uniprot_id]["mappings"]
        
        # store the chain ids that correspond to this UniProt ID
        uniprot_chain_mapping_dict[uniprot_id] = []
        
        for mapping in mappings:
            entity_id = mapping["entity_id"]
            
            chain_id = mapping["chain_id"]
            uniprot_chain_mapping_dict[uniprot_id].append(chain_id)
            
            pdb_start = mapping["start"]["residue_number"]
            pdb_end = mapping["end"]["residue_number"]
            uniprot_start = mapping["unp_start"]
            uniprot_end = mapping["unp_end"]
        
    # "flip" the uniprot > pdb chain mapping
    # get all unique chain IDs
    chain_IDs = (list(set([item for sublist in 
                           uniprot_chain_mapping_dict.values() for item 
                           in sublist])))

    chain_uniprot_mapping_dict = {}
    # iterate through every chain ID
    for chain_i in chain_IDs:

        # iterate through every uniprot ID and check to 
        # see if chain ID is mapped
        chain_uniprot_mapping_dict[chain_i] = []
        for uniprot_i in uniprot_chain_mapping_dict.keys():
            if chain_i in uniprot_chain_mapping_dict[uniprot_i]:
                chain_uniprot_mapping_dict[chain_i].append(uniprot_i)
                
    return chain_uniprot_mapping_dict

def PDB_chains_to_uniprot(interacting_chains_list, 
                          chain_to_UniProt_mapping_dict):
    '''
    Get interacting chains from PDB structure mapped to UniProt IDs.
    
    This function takes the list of interacting chains from function
    get_interacting_chains_from_PDB() and the chain to UniProt mappings
    from function list_uniprot_pdb_mappings() and returns a list of 
    interacting chains using UniProt IDs.
    
    Parameters
    ----------
    Interacting Chains: list
    Chain to UniProt Map: dict
    
    Returns
    -------
    Interacting Chains
        List of interacting chains using UniProt IDs
    
    Examples
    --------
    # (1) Obtain list of interacting chains from 6YW7 structure
    # (2) Obtain a mapping of PDB ID 6YW7 chains to UniProt IDs
    # (3) Obtain list of interacting chains from 6YW7 
    #     structure using UniProt IDs 
    >>> interacting_chains_list = get_interacting_chains_from_PDB('6YW7', '.', 6)
    Downloading PDB structure '6YW7'...
    >>> chain_to_UniProt_mapping_dict = list_uniprot_pdb_mappings('6YW7') 
    >>> interacting_UniProt_IDs = PDB_chains_to_uniprot(interacting_chains_list, chain_to_UniProt_mapping_dict)
    >>> sorted(interacting_UniProt_IDs)
    [['O15144', 'P59998'], ['O15511', 'Q92747'], ['P59998', 'O15511'], ['P59998', 'Q92747'], ['P61158', 'O15144'], ['P61158', 'O15145'], ['P61158', 'P61160'], ['P61160', 'O15511'], ['P61160', 'P59998']]
    '''
    interacting_UniProt_IDs = []
    for interacting_chain_pair_i in interacting_chains_list:

        # get UniProt IDs that map to each chain ID
        chain_i_UniProts = (
            chain_to_UniProt_mapping_dict[interacting_chain_pair_i[0]])
        chain_j_UniProts = (
            chain_to_UniProt_mapping_dict[interacting_chain_pair_i[1]])

        # store every pair of UniProt IDs that correspond 
        # to the interacting chains
        for chain_i_UniProt_ID in chain_i_UniProts:
            for chain_j_UniProt_ID in chain_j_UniProts:
                interacting_UniProt_IDs.append(
                    [chain_i_UniProt_ID,chain_j_UniProt_ID])
                
    return interacting_UniProt_IDs

def PDB_to_interacting_chains_uniprot_maps(PDB_ID, 
                                           protein_structure_dir, 
                                           interact_dist_threshold):
    '''
    Get interacting chains from PDB structure mapped to UniProt IDs and 
    PDB chain to UniProt mappings.
    
    This is a wrapper function for functions 
    (1) get_interacting_chains_from_PDB(),
    (2) list_uniprot_pdb_mappings(), and
    (3) PDB_chains_to_uniprot() 
    to get a list of interacting chains from PDB structure using UniProt labels 
    and the chain-to-UniProt mapping for this PDB structure.
    
    Parameters
    ----------
    PDB ID: str
    directory to store PDB file: str
    distance threshold: int
    
    Returns
    -------
    Chain to UniProt Map
        Dictionary of PDB ID chain to UniProt ID mappings
    Interacting Chains
        List of interacting chains using UniProt IDs
    
    Examples
    --------
    >>> (chain_to_UniProt_mapping_dict, interacting_UniProt_IDs) = PDB_to_interacting_chains_uniprot_maps('6NMI', '.', 6)
    Downloading PDB structure '6NMI'...
    '''
    # get list of chain pairs that interact in PDB structure 
    interacting_chains_list = get_interacting_chains_from_PDB(PDB_ID, 
                                                    protein_structure_dir, 
                                                    interact_dist_threshold)
    
    # get chain > UniProt ID mappings for this PDB structure
    chain_to_UniProt_mapping_dict = list_uniprot_pdb_mappings(PDB_ID)
    
    # get list of chains pairs that interact in PDB structure using UniProts
    interacting_UniProt_IDs = PDB_chains_to_uniprot(interacting_chains_list, 
                                                chain_to_UniProt_mapping_dict)

    return [chain_to_UniProt_mapping_dict, interacting_UniProt_IDs]